use HLL;

class TinyLang::Grammar is HLL::Grammar {
    rule TOP {
        <.newblock>
        <EXPR> ** ';'
    }
    token newblock { <?> }
    
    proto token term { <...> }
    token term:sym<var> { <variable> <![(]> }
    token term:sym<decl> { <variable_declaration> }
    token term:sym<call> { <ident> '(' <arglist> ')' }
    token term:sym<value> { <value> }
    
    rule arglist { <EXPR> ** ',' }
    
    rule variable_declaration { 'var' <variable> }
    token variable { <ident> }
    
    proto token value { <...> }
    token value:sym<string> { <?["]> <quote_EXPR> }
    
    ## Operators
    INIT {
        # Precedence levels.
        TinyLang::Grammar.O(':prec<z=>, :assoc<left>',  '%concatenation');
        TinyLang::Grammar.O(':prec<x=>, :assoc<right>', '%assignment');
    }
    token infix:sym<~> { <sym> <O('%concatenation')> }
    token infix:sym<=> { <sym> <O('%assignment')> }
}

class TinyLang::Actions is HLL::Actions {
    my @BLOCK;
    
    method TOP($/) {
        my $program := @BLOCK.shift;
        for $<EXPR> {
            $program.push($_.ast);
        }
        make $program;
    }
    
    method newblock($/) {
        @BLOCK.unshift(PAST::Block.new());
    }
    
    method term:sym<var>($/) {
        make $<variable>.ast;
    }
    
    method term:sym<decl>($/) {
        make $<variable_declaration>.ast;
    }
    
    method term:sym<call>($/) {
        my $call := $<arglist>.ast;
        $call.pasttype('call');
        $call.name(~$<ident>);
        make $call;
    }
    
    method term:sym<value>($/) {
        make $<value>.ast;
    }
    
    method arglist($/) {
        my $args := PAST::Op.new();
        for $<EXPR> {
            $args.push($_.ast);
        }
        make $args;
    }
    
    method variable_declaration($/) {
        # Put in symbol table as well as making var node.
        my $name := ~$<variable>;
        @BLOCK[0].symbol($name, :scope('lexical'));
        make PAST::Var.new( :name($name), :isdecl(1) );
    }
    
    method variable($/) {
        make PAST::Var.new( :name(~$<ident>) );
    }
    
    method value:sym<string>($/) {
        make $<quote_EXPR>.ast;
    }
    
    method infix:sym<~>($/) {
        make PAST::Op.new( :pirop('concat Sss') );
    }
    
    method infix:sym<=>($/) {
        make PAST::Op.new( :pasttype('bind') );
    }
}

class TinyLang::Compiler is HLL::Compiler {
}

sub MAIN(*@ARGS) {
    # Create and configure compiler object.
    my $tlcomp := TinyLang::Compiler.new();
    $tlcomp.language('tinylang');
    $tlcomp.parsegrammar(TinyLang::Grammar);
    $tlcomp.parseactions(TinyLang::Actions);

    # Enter the compiler.
    $tlcomp.command_line(@ARGS, :encoding('utf8'));
}

# Built-in subs just come here. :-)
our sub writeline($message) {
    say($message);
}
